all
rule 'MD009', :br_spaces => 2
rule 'MD013', :line_length => 120, :tables => false, :code_blocks => false
rule 'MD029', :style => :ordered
exclude_rule 'MD033'
exclude_rule 'MD034'
exclude_rule 'MD041'
