/*
    SPDX-FileCopyrightText: Lieven Hey <lieven.hey@kdab.com>
    SPDX-FileCopyrightText: Milian Wolff <milian.wolff@kdab.com>
    SPDX-FileCopyrightText: 2016 Klarälvdalens Datakonsult AB, a KDAB Group company, info@kdab.com

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "perfoutputwidget.h"

PerfOutputWidget::PerfOutputWidget(QWidget* parent)
    : QWidget(parent)
{
}

PerfOutputWidget::~PerfOutputWidget() = default;
